from .preprocess import *
from .data_loading import *
from .basicNetwork import *
from .string_manipulate import *
from .opt_gaussian import *
from .data_stats import *
from .IO import *
from .dependency import *
from .DManager import *
from .imputation import *
from .type_check import *
from .kde import *
from .L2X import *
from .gumbel import *
from .run_k_fold_on_function import *
from .classifiers import *
from .sampling import *
from .distribution_modeling import *
from .operations import *
from .matrix_manipulate import *
from .explainer import *
from .error import *



#!/usr/bin/env python


import sys
import matplotlib
import numpy as np
import random
import itertools
import socket
import sklearn.metrics
from scipy.optimize import minimize
from scipy.optimize import Bounds
from sklearn import preprocessing
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.random import sample_without_replacement

np.set_printoptions(precision=4)
np.set_printoptions(threshold=sys.maxsize)
np.set_printoptions(linewidth=300)
np.set_printoptions(suppress=True)



class opt_gaussian():
	def __init__(self, X, Y, Y_kernel):	#	X=data, Y=label, σ_type='ℍ', or 'maxKseparation'
		ń = X.shape[0]
		ð = X.shape[1]
		self.Y_kernel = Y_kernel

		if ń > 200:
			#	Down sample first
			samples = sample_without_replacement(n_population=ń, n_samples=200)
			X = X[samples,:]
			Y = Y[samples]
			ń = X.shape[0]


		if Y_kernel == 'linear':
			self.Ⅱᵀ = np.ones((ń,ń))
			ńᒾ = ń*ń
			Yₒ = OneHotEncoder(categories='auto', sparse=False).fit_transform(np.reshape(Y,(len(Y),1)))
			self.Kᵧ = Kᵧ = Yₒ.dot(Yₒ.T)
			ṉ = np.sum(Kᵧ)
			σᵧ = 1

			HKᵧ = self.Kᵧ - np.mean(self.Kᵧ, axis=0)								# equivalent to Γ = Ⲏ.dot(Kᵧ).dot(Ⲏ)
			self.Γ = HKᵧH = (HKᵧ.T - np.mean(HKᵧ.T, axis=0)).T

		elif Y_kernel == 'Gaussian':
			Ðᵧ = sklearn.metrics.pairwise.pairwise_distances(Y)
			σᵧ = np.median(Ðᵧ)
			self.Ðᵧᒾ = (-Ðᵧ*Ðᵧ)


		#	X_kernel == 'Gaussian'
		Ðₓ = sklearn.metrics.pairwise.pairwise_distances(X)
		σₓ = np.median(Ðₓ)
		self.Ðₓᒾ = (-Ðₓ*Ðₓ)

		self.σ = [σₓ, σᵧ]

	def minimize_H(self):
		self.result = minimize(self.ℍ, self.σ, method='L-BFGS-B', options={'gtol': 1e-5, 'disp': False}, bounds=Bounds(0.05, 100000))


	def ℍ(self, σ, X=None, Y=None):
		[σₓ, σᵧ] = σ
		if X is None:
			Kₓ = np.exp(self.Ðₓᒾ/(2*σₓ*σₓ))
			if self.Y_kernel == 'linear':
				Γ = self.Γ		
			elif self.Y_kernel == 'Gaussian':
				Kᵧ = np.exp(self.Ðᵧᒾ/(σᵧ*σᵧ))
				HKᵧ = Kᵧ - np.mean(Kᵧ, axis=0)								# equivalent to Γ = Ⲏ.dot(Kᵧ).dot(Ⲏ)
				Γ = HKᵧH = (HKᵧ.T - np.mean(HKᵧ.T, axis=0)).T
		else:
			Ðₓ = sklearn.metrics.pairwise.pairwise_distances(X)
			Ðₓᒾ = (-Ðₓ*Ðₓ)
			Kₓ = np.exp(Ðₓᒾ/(2*σₓ*σₓ))

			ń = X.shape[0]
			ð = X.shape[1]

			Yₒ = OneHotEncoder(categories='auto', sparse=False).fit_transform(np.reshape(Y,(len(Y),1)))
			Kᵧ = Yₒ.dot(Yₒ.T)

			HKᵧ = Kᵧ - np.mean(Kᵧ, axis=0)								# equivalent to Γ = Ⲏ.dot(Kᵧ).dot(Ⲏ)
			Γ = HKᵧH = (HKᵧ.T - np.mean(HKᵧ.T, axis=0)).T

		loss = -np.sum(Kₓ*Γ)
		return loss


def get_opt_σ(X,Y, Y_kernel='linear', compute_final_hsic=False):
	optimizer = opt_gaussian(X,Y, Y_kernel=Y_kernel)
	optimizer.minimize_H()

	if compute_final_hsic: 
		optimizer.final_hsic = -optimizer.ℍ(optimizer.result.x, X=X, Y=Y)
	return optimizer

def get_opt_σ_via_random(X,Y, Y_kernel='Gaussian'):
	optimizer = opt_gaussian(X,Y, Y_kernel=Y_kernel)

	opt = 0
	opt_σ = 0
	for m in range(1000):
		σ = (7*np.random.rand(2)).tolist()
		new_opt = -optimizer.ℍ(σ)
		if opt < new_opt:
			opt = new_opt
			opt_σ = σ

	print('Random Result ')
	print('\tbest_σ : ', opt_σ)
	print('\tmax_HSIC : ' , opt)


if __name__ == "__main__":
	#data_name = 'wine'
	data_name = 'spiral_arm'
	X = np.loadtxt('../dataset/' + data_name + '.csv', delimiter=',', dtype=np.float64)			
	Y = np.loadtxt('../dataset/' + data_name + '_label.csv', delimiter=',', dtype=np.int32)			
	X = preprocessing.scale(X)


	optimizer = get_opt_σ(X,Y, compute_final_hsic=True)
	best_σ = optimizer.result.x
	max_HSIC = optimizer.final_hsic
	print('Optimized Result ')
	print('\tbest_σ [σₓ, σᵧ]: ', best_σ)
	print('\tmax_HSIC : ' , max_HSIC)


	optimized_results = get_opt_σ_via_random(X,Y, Y_kernel='linear')







import numpy as np
import wuml
import torch
from torch.autograd import Variable
import torch.nn.functional as f

#torch.autograd.set_detect_anomaly(True)

class l2x:
	def __init__(self, data, max_epoch=2000, learning_rate=0.001, data_imbalance_weights=None,
					X_dataType=torch.FloatTensor, Y_dataType=torch.FloatTensor, regularizer_weight=0,
					selector_network_structure=None, predictor_network_structure=None, 
					pre_trained_file=None, use_binary=True):
		self.data = data
		self.d = d = data.shape[1]
		self.max_epoch = max_epoch
		self.lr = learning_rate
		self.W = data_imbalance_weights
		self.X_dataType = X_dataType
		self.Y_dataType = Y_dataType
		self.regularizer_weight = regularizer_weight
		self.use_binary = use_binary
		self.pre_trained_file = pre_trained_file

		if pre_trained_file is None:
			if selector_network_structure is None:
				#self.θˢ = wuml.flexable_Model(d, [(d,'relu'),(1000,'relu'),(1000,'relu'),(d*2,'none')])
				self.θˢ = wuml.flexable_Model(d, [(d,'relu'),(1000,'relu'),(1000,'relu'),(d*d,'none')])
			else:
				self.θˢ = wuml.flexable_Model(d, selector_network_structure)
	
			if predictor_network_structure is None:
				self.θᴼ = wuml.flexable_Model(d, [(d,'relu'),(1000,'relu'),(1000,'relu'),(1,'none')])
			else:
				self.θᴼ = wuml.flexable_Model(d, predictor_network_structure)
		else:
			θ = wuml.pickle_load(pre_trained_file)
			self.θˢ = θ['θˢ']
			self.θᴼ = θ['θᴼ']

		if torch.cuda.is_available(): 
			self.device = 'cuda'
			self.θˢ.to(self.device)		# store the network weights in gpu or cpu device
			self.θᴼ.to(self.device)		# store the network weights in gpu or cpu device
		else: self.device = 'cpu'
		self.info()

	def export_network(self, save_path):
		network = {}
		network['θᴼ'] = self.θᴼ
		network['θˢ'] = self.θˢ

		wuml.pickle_dump(network, save_path)


	def info(self, save_path=None):
		using_weights = True
		if self.W is None: using_weights = False

		txt = 'L2X Regression\n'
		txt += '\tDevice: %s\n'%(self.device)
		txt += '\tBatch size: %d\n'%(self.data.batch_size)
		txt += '\tData Dimension: %d x %d\n'%(self.data.shape[0], self.data.shape[1])
		txt += '\tUse Binary One Hot: %r\n'%(self.use_binary)
		txt += '\tUse Pre-train File: %s\n'%(str(self.pre_trained_file))
		txt += '\tMax num of Epochs: %d\n'%(self.max_epoch)
		txt += '\tInitial Learning Rate: %.7f\n'%(self.lr)
		txt += '\tUsing weighted Samples: %r\n'%(using_weights)
		txt += '\tθˢ structure: %s\n'%(str(self.θˢ.networkStructure))
		txt += '\tθᴼ structure: %s\n'%(str(self.θᴼ.networkStructure))
		print(txt)
		#save_path


	def get_Selector(self, x, round_result=False, test_phase=False):
		x = wuml.ensure_tensor(x)
		[d, device] = [self.d, self.device]

		if test_phase:
			if self.use_binary:
				xᵃ = self.θˢ(x)
				xᵇ = xᵃ.view(-1, d, 2)	# reshape the data into num_of_samples x data dimension x num_of_groups 
				xᶜ = torch.nn.Softplus()(xᵇ)

				S = torch.argmin(xᶜ, dim=2)
			else:
				xᵃ = self.θˢ(x)
				q = int(xᵃ.shape[1]/d)
				xᵇ = xᵃ.view(-1, q, d)	# reshape the data into num_of_samples x data dimension x num_of_groups 
				xᶜ = torch.nn.Softplus()(xᵇ)
				xᶜ = f.normalize(xᶜ, p=1, dim=2)

				[S,ind] = torch.max(xᶜ, dim=1)
		else:
			if self.use_binary:
				xᵃ = self.θˢ(x)
				xᵇ = xᵃ.view(-1, d, 2)	# reshape the data into num_of_samples x data dimension x num_of_groups 
				xᶜ = torch.nn.Softplus()(xᵇ)
				xᶜ = f.normalize(xᶜ, p=1, dim=2)
	
				xᵈ = wuml.gumbel(xᶜ, device=device)		# Group matrix
				S = xᵈ[:,:,0]
			else:
				xᵃ = self.θˢ(x)
				q = int(xᵃ.shape[1]/d)
				xᵇ = xᵃ.view(-1, q, d)	# reshape the data into num_of_samples x data dimension x num_of_groups 
	
				xᶜ = torch.nn.Softplus()(xᵇ)
				xᶜ = f.normalize(xᶜ, p=1, dim=2)
				xᵈ = wuml.gumbel(xᶜ, device=device)		# Group matrix
				
				[S,ind] = torch.max(xᵈ, dim=1)





		if torch.isnan(S[0,0]): 
			print('\n\nNan detected within get_Selector function in L2X')
			import pdb; pdb.set_trace() 

		if round_result: S = torch.round(S)
		return S

	def __call__(self, x):
		x = wuml.ensure_tensor(x)

		S = self.get_Selector(x, test_phase=True)
		x̂ = x*S
		ŷ = self.θᴼ(x̂)
		return ŷ

	def loss(self, x, y, ind):		# this function is not working currently
		[n, λ] = [len(ind), self.regularizer_weight]

		if self.W is None:
			W = torch.ones(n, 1)
			W = Variable(W.type(self.X_dataType), requires_grad=False)
		else:
			W = self.W[ind]
			W = torch.from_numpy(W)
			W = Variable(W.type(self.X_dataType), requires_grad=False)

		W = W.to(self.device, non_blocking=True )
		W = torch.squeeze(W)

		S = self.get_Selector(x)
		x̂ = x*S
		ŷ = self.θᴼ(x̂)
		ŷ = torch.squeeze(ŷ)

		if λ == 0:
			regularizer = λ*torch.sum(S)
			loss = torch.sum(W*((y - ŷ)** 2))/n + regularizer
		else:
			loss = torch.sum(W*((y - ŷ)** 2))/n

		#if loss < 0.01:
		#	import pdb; pdb.set_trace()

		if torch.isnan(loss): import pdb; pdb.set_trace()
		return loss

	def on_new_epoch_call_back(self, loss_avg, epoch, lr):
		pass

		#if epoch%2 == 0:
		#	for param in self.θˢ.parameters(): 
		#		print(param[0:10,0:10], '\n')
		#		break

			#for param in self.θᴼ.parameters(): 
			#	print(param[0:10,0:10])
			#	break

		#	S = self.get_Selector(self.data.X, round_result=True)
		#	import pdb; pdb.set_trace()

	def train(self, print_status=True):
		trainLoader = self.data.get_data_as('DataLoader')
		param = list(self.θˢ.parameters()) + list(self.θᴼ.parameters())
		[ℓ, TL, mE, Dev] = [self.loss, trainLoader, self.max_epoch, self.device]
		[Xtype, Ytype] = [self.X_dataType, self.Y_dataType]

		wuml.run_SGD(ℓ, param, TL, Dev, lr=self.lr, 
					max_epoch=mE, X_dataType=Xtype, Y_dataType=Ytype,
					on_new_epoch_call_back=self.on_new_epoch_call_back)






import wuml
import numpy as np
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import LabelEncoder


def sort_matrix_rows_by_a_column(X, column_names):
	df = wuml.ensure_DataFrame(X)
	sorted_df = df.sort_values(by=column_names)

	return sorted_df



def one_hot_encoding(Y, output_data_type='same'): 
	'''
		output_data_type : 'same' implies to use the same data type as input
							else use the data type you want, i.e., 'ndarray' , 'Tensor', 'DataFrame'
	'''
	dtype = wuml.data_type(Y)
	Y1 = wuml.ensure_numpy(Y)

	Y1 = np.reshape(Y1,(len(Y1),1))
	Yₒ = OneHotEncoder(categories='auto', sparse=False).fit_transform(Y1)

	if output_data_type == 'same':
		Yₒ = wuml.ensure_data_type(Yₒ, type_name=dtype)
	else:
		Yₒ = wuml.ensure_data_type(Yₒ, type_name=output_data_type)

	return Yₒ

def one_hot_to_label(Yₒ):
	Y = np.argmax(Yₒ, axis=1)
	return Y


Y = np.array([0,0,1,1,2,2,3,3])
Yₒ = one_hot_encoding(Y)
Y2 = one_hot_to_label(Yₒ)




from sklearn import preprocessing
import sys
import wuml
import pandas as pd
import torch
import numpy as np
from sklearn.preprocessing import LabelEncoder
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader

class wData:
	def __init__(self, xpath=None, ypath=None, column_names=None, label_column_name=None, dataFrame=None, 
					X_npArray=None, Y_npArray=None, row_id_with_label=None, sample_id_included=False, 
					label_type=None, #it should be either 'continuous' or 'discrete'
					torchDataType=torch.FloatTensor, batch_size=20, columns_to_ignore=None):
		'''
			row_id_with_label :  None = no labels, 0 = top row is the label
			dataFrame: if dataFrame is set, it ignores the path and use the dataFrame directly as the data itself
			ypath: loads the data as the label
			label_column_name: if the label is loaded together with xpath, this separates label into Y
		'''
		self.label_column_name = label_column_name

		if dataFrame is not None:
			self.df = dataFrame
		elif X_npArray is not None:
			self.df = pd.DataFrame(X_npArray, columns=column_names)
		else:
			self.df = pd.read_csv (xpath, header=row_id_with_label, index_col=False)

		if type(self.df.columns).__name__ == 'str': 	# if text column names, strip away white space
			self.df.columns = self.df.columns.str.replace(' ','')

		self.Y = None
		if Y_npArray is not None:
			self.Y = Y_npArray
		elif ypath is not None: 
			if label_type is None: raise ValueError('If you are using labels, you must include the argument label_type= "continuout" or "discrete"')
			self.Y = np.loadtxt(ypath, delimiter=',', dtype=np.float32)			
			if label_type == 'discrete': self.Y = LabelEncoder().fit_transform(self.Y)	#Make sure label start from 0
		elif label_column_name is not None:
			if label_type is None: raise ValueError('If you are using labels, you must include the argument label_type= "continuout" or "discrete"')
			self.Y = self.df[label_column_name].values
			if label_type == 'discrete': self.Y = LabelEncoder().fit_transform(self.Y)	#Make sure label start from 0
			self.delete_column(label_column_name)

		if columns_to_ignore is not None: self.delete_column(columns_to_ignore)

		self.X = self.df.values
		self.batch_size = batch_size
		self.shape = self.df.shape
		self.torchDataType = torchDataType				
		self.torchloader = None
		self.label_type = label_type

		if torch.cuda.is_available(): self.device = 'cuda'
		else: self.device = 'cpu'

	def delete_column(self, column_name):
		if type(column_name) == type([]):
			for name in column_name:
				if name in self.df.columns:
					del self.df[name]

		elif type(column_name) == type(''):
			if column_name in self.df.columns:
				del self.df[column_name]

	def info(self):
		print(self.df.info())

	def get_data_as(self, data_type): #'DataFrame', 'read_csv', 'Tensor'
		if data_type == 'wData': return self
		if data_type == 'Tensor': 
			x = torch.from_numpy(self.df.values)
			x = Variable(x.type(self.torchDataType), requires_grad=False)
			X = x.to(self.device, non_blocking=True )
			return X
		if data_type == 'DataFrame': return self.df
		if data_type == 'ndarray': 
			#self.df.values[subset]
			return self.df.values
		if data_type == 'DataLoader':		# and self.torchloader is None 
			self.DM = wuml.DManager(self.df.values, self.Y)
			self.torchloader = DataLoader(dataset=self.DM, batch_size=self.batch_size, shuffle=True, pin_memory=True, num_workers=1)
			return self.torchloader

	def to_csv(self, path, add_row_indices=False, include_column_names=True, float_format='%.4f'):
		LCn = self.label_column_name

		if LCn is not None:
			self.df[LCn] = self.Y
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format)
			self.delete_column(LCn)
		elif self.Y is not None:
			self.df['label'] = self.Y
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format )
			self.delete_column('label')
		else:
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format)

	def __getitem__(self, item):
		return self.df.values[item]

	def __str__(self): 
		return str(self.df)
 




import numpy as np
import random
import string

def gen_random_string(str_len=5, letters=string.ascii_letters):
	#letters = string.ascii_lowercase
	#letters = string.ascii_letters
	#letters = string.ascii_uppercase
	#letters = string.digits
	#letters = string.punctuation
	return ''.join(random.choice(letters) for i in range(str_len)) 

def gen_exponential(λ=1, size=1):
	# λ exp(-λx) , μ=1/λ
	return np.random.exponential(scale=1/λ, size=size)


#category_range	: if 5, it will select from 0,1,2,3,4
#n				: number of samples returned	
#probabilities 	: the probability of each instance
#example		: np.random.choice(5, 3, p=[0.1, 0, 0.3, 0.6, 0])
def gen_categorical(category_range, n, probabilities):
	return np.random.choice(category_range, n, p=probabilities)
	




from sklearn.svm import SVC
import numpy as np

class svm:
	def __init__(self, X, Y, kernel='rbf'):
		if len(Y.shape) == 2: Y = np.squeeze(Y)
		self.SVM = SVC(kernel=kernel)
		self.SVM.fit(X, Y)

	def __call__(self, X):
		return self.SVM.predict(X)



#!/usr/bin/env python

#from distutils.core import setup
from setuptools import setup
from os import path
import os

#def read_file(filename):
#	with open(filename) as file:
#		return file.read()

#def read_file(filename):
#	fn = os.path.join(os.path.dirname(__file__), filename)
#	if os.path.exists(fn):
#		with open(fn) as file:
#			return file.read()
#	else:
#		return 'Basic ML Library'

#this_directory = path.abspath(path.dirname(__file__))
#with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
#	long_description = f.read()



setup(
  name = 'wuml',         # How you named your package folder (MyLib)
  packages = ['wuml'],   # Chose the same as "name"
  version = '0.062',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'A library that simplifies some basic ML stuff.',   # Give a short description about your library
  #long_description=read_file('README.md'),
  long_description='https://github.com/endsley/wuML',
  long_description_content_type='text/markdown',  
  author = 'Chieh Wu',                   # Type in your name
  author_email = 'chieh.t.wu@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/endsley/wuML.git',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/endsley/wuML/archive/refs/tags/0.062.tar.gz',    # I explain this later on
  keywords = ['ML', 'data analysis'],   # Keywords that define your package best
  include_package_data=False,
  install_requires=[            # I get to this in a second
          'matplotlib',
          'wplotlib',
          'sklearn',
          'shap',
          'numpy',
          'pandas',
          'torch',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
)




import numpy as np
import wuml
import torch
import pandas as pd
from torch.autograd import Variable


def data_type(data):
	return type(data).__name__

def ensure_data_type(data, type_name='ndarray'):

	if type_name=='ndarray':
		return ensure_numpy(data)
	elif type_name=='DataFrame':
		return ensure_DataFrame(data)
	elif type_name=='Tensor':
		return ensure_tensor(data)
	elif type_name=='wData':
		return ensure_wData(data)

def ensure_wData(data, column_names=None):
	if type(data).__name__ == 'ndarray': 
		return wuml.wData(X_npArray=data, column_names=column_names)
	elif type(data).__name__ == 'wData': 
		return data
	elif type(data).__name__ == 'DataFrame': 
		return wuml.wData(dataFrame=data, column_names=column_names)
	elif type(data).__name__ == 'Tensor': 
		X = data.detach().cpu().numpy()
		return wuml.wData(X_npArray=X, column_names=column_names)


def ensure_DataFrame(data, columns=None):
	if type(data).__name__ == 'ndarray': 
		df = pd.DataFrame(data)
	elif type(data).__name__ == 'wData': 
		df = data.df
		df.columns = data.df.columns
	elif type(data).__name__ == 'DataFrame': 
		df = data
	elif type(data).__name__ == 'Tensor': 
		X = data.detach().cpu().numpy()
		df = pd.DataFrame(X)
	
	if columns is not None:
		df.columns = columns

	return df


def ensure_numpy(data, rounding=None):
	if type(data).__name__ == 'ndarray': 
		if len(data.shape) == 1:
			X = np.atleast_2d(data).T
		else:
			X = data
	elif type(data).__name__ == 'wData': 
		X = data.df.values
	elif type(data).__name__ == 'DataFrame': 
		X = data.values
	elif type(data).__name__ == 'Tensor': 
		X = data.detach().cpu().numpy()
	elif np.isscalar(data):
		X = np.array([[data]])
	else:
		raise ValueError('Unknown dataType %s'%type(data).__name__)

	if rounding is not None: X = np.round(X, rounding)
	return X


def ensure_tensor(data, dataType=torch.FloatTensor):
	if torch.cuda.is_available(): 
		device = 'cuda'
	else: self.device = 'cpu'

	if type(data).__name__ == 'ndarray': 
		x = torch.from_numpy(data)
		x = Variable(x.type(dataType), requires_grad=False)
		X = x.to(device, non_blocking=True )
	elif type(data).__name__ == 'wData': 
		X = data.get_data_as('Tensor')
	elif type(data).__name__ == 'DataFrame': 
		X = data.values
	elif np.isscalar(data):
		X = np.array([[data]])
	elif type(data).__name__ == 'Tensor': 
		x = Variable(data.type(dataType), requires_grad=False)
		X = x.to(device, non_blocking=True )
	else:
		raise ValueError('Unknown dataType %s'%type(data).__name__)

	return X





